/*  WB-tree File Based Associative String Data Base System. */
/*  Copyright (c) 1991, 1992, 1993, 2000 Free Software Foundation, Inc. */
/*  */
/*  This program is free software; you can redistribute it and/or modify */
/*  it under the terms of the GNU General Public License as published by */
/*  the Free Software Foundation; either version 2, or (at your option) */
/*  any later version. */
/*  */
/*  This program is distributed in the hope that it will be useful, */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the */
/*  GNU General Public License for more details. */
/*  */
/*  You should have received a copy of the GNU General Public License */
/*  along with this software; see the file COPYING.  If not, write to */
/*  the Free Software Foundation, 59 Temple Place, Suite 330, Boston, */
/*  MA 02111, USA. */
/*  */
/*  As a special exception, the Free Software Foundation gives permission */
/*  for additional uses of the text contained in its release of GUILE. */
/*  */
/*  The exception is that, if you link the GUILE library with other files */
/*  to produce an executable, this does not by itself cause the */
/*  resulting executable to be covered by the GNU General Public License. */
/*  Your use of that executable is in no way restricted on account of */
/*  linking the GUILE library code into it. */
/*  */
/*  This exception does not however invalidate any other reasons why */
/*  the executable file might be covered by the GNU General Public License. */
/*  */
/*  This exception applies only to the code released by the */
/*  Free Software Foundation under the name GUILE.  If you copy */
/*  code from other Free Software Foundation releases into a copy of */
/*  GUILE, as the General Public License permits, the exception does */
/*  not apply to the code that you add in this way.  To avoid misleading */
/*  anyone as to the status of such modified files, you must delete */
/*  this exception notice from them. */
/*  */
/*  If you write modifications of your own for GUILE, it is your choice */
/*  whether to permit this exception to apply to your modifications. */
/*  If you do not wish that, delete this exception notice. */

/*  routines in this file return success codes */


#include "wbsys.h"
/* @section SEGs */
/* @cindex SEG */
/*  Segment not in use? */

int seg_free_P(seg)
     int seg;
{
  if (!(lck_tab))
    init_wb(0x4b, 0x96, 0x800);
  if ((0 > (seg))
      || ((seg)>=(num_segs)))
    {
      dprintf((diagout, ">>>>ERROR<<<< bad segment number %d\n", seg));
      return 0;
    }
  else if ((!(seg_port(seg)))
      && (!(seg_str(seg)))
      && (!(seg_used(seg))))
    return !0;
  else return 0;
}

/* TBD - need to lck seg here. */
/*  Segment will be read-only if MODE is #f. */

/*  */
/* @var{seg} should be a non-negative number.  Opens the database @var{seg} in @var{filename} and */
/* returns @var{seg} if successful, a status code otherwise.  The database @var{seg} */
/* will be read-only if the @var{mode} argument is 0.  It will be read-write if */
/* the @var{mode} argument is 2. */

int open_seg(seg, filename, mode)
     int seg;
     unsigned char *filename;
     int mode;
{
  int bsiz = 0;
  unsigned char *reason_str;
  if (!(mode))
    mode = 0;
  if (!(seg_free_P(seg)))
    {
      dprintf((diagout, ">>>>ERROR<<<< %s: segment in use %d\n", "open_seg", seg));
      return argerr;
    }
  else if ((bsiz = min_file_blk_size(filename),bsiz = max(3*0x80, bsiz),(bsiz)>(blk_size)))
    {
      dprintf((diagout, ">>>>ERROR<<<< unsupported %s %d > %d\n", "bsiz", bsiz, blk_size));
      return argerr;
    }
  else {
    int file = mode
      ?blk_file_open_modify(filename, bsiz)
      :blk_file_open_read_only(filename, bsiz);
L_loop:
    if (mode
	  ?output_port_P(file)
	  :input_port_P(file))
      {
	seg_set_port(seg, file);
	seg_set_str(seg, filename);
	seg_set_used(seg, 2);
	seg_set_bsiz(seg, bsiz);
	seg_set_flc_len(seg, mode
	    ?-1
	    :-2);
	{
	  HAND *han = seg_rt_han(seg);
	  unsigned char tmp_str[5];
	  if (err_P(bt_open(seg, 0, han, (wcb_sap)+(wcb_sar))))
	    {
	      reason_str = "bt-open 0";
	      goto L_errout;
	    }
	  else if (2!=(bt_get(han, "BSIZ", 4, tmp_str)))
	    {
	      reason_str = "BSIZ";
	      goto L_errout;
	    }
	  else if ((bsiz)!=(str2short(tmp_str, 0)))
	    {
	      blk_file_close(file);
	      bsiz = str2short(tmp_str, 0);
	      if ((bsiz)>(blk_size))
		{
		  reason_str = "BSIZ too big.";
		  goto L_errout;
		}
	      else {
		file = mode
		  ?blk_file_open_modify(filename, bsiz)
		  :blk_file_open_read_only(filename, bsiz);
		goto L_loop;
	      }
	    }
	  else if (4!=(bt_get(han, "USED", 4, tmp_str)))
	    {
	      reason_str = "USED";
	      goto L_errout;
	    }
	  else {
	    seg_set_used(seg, str2long(tmp_str, 0));
	    if (5!=(bt_get(han, "FLD", 3, tmp_str)))
	      {
		reason_str = "FLD";
		goto L_errout;
	      }
	    else if (err_P(bt_open(seg, str2long(tmp_str, 1), seg_fl_han(seg), wcb_sar)))
	      {
		reason_str = "FLC";
		goto L_errout;
	      }
	    else {
	      if ((han_typ(seg_fl_han(seg)))!=(frl_typ))
		dprintf((diagout, "Older type freelist - still supported.\n"));
	      han_set_wcb(seg_fl_han(seg), wcb_sar);
	      return seg;
	    }
	  }
	}
      }
    else {
      if (mode
	    ?input_port_P(file)
	    :output_port_P(file))
	blk_file_close(file);
      dprintf((diagout, ">>>>ERROR<<<< could not open file %s\n", filename));
      return ioerr;
    }
  }
L_errout:
  dprintf((diagout, ">>>>ERROR<<<< not a database %s %s\n", filename, reason_str));
  blk_file_close(seg_port(seg));
  seg_set_port(seg, 0);
  seg_set_str(seg, 0);
  seg_set_used(seg, 0);
  return typerr;
}

/*  */
/* Closes database segment @var{seg} and the file containing it.  If @var{hammer} is */
/* NULL, then if there are any problems freeing buffers, then the close */
/* is aborted.  A status code is returned. */

int close_seg(seg, hammer)
     int seg;
     int hammer;
{
  if ((!(seg_str(seg)))
      || (!(seg_used(seg))))
    {
      dprintf((diagout,
	 ">>>>ERROR<<<< %s: segment %d already closed\n",
	 "close_seg",
	 seg));
      return argerr;
    }
  else {
    flush_flc(seg, 5);
    /*leave only enough blocks to fit in flc in superblock.*/
    if ((seg_flc_len(seg))>=0)
      {
	unsigned char tmp_str[0x14];
	{
	  int i = -1+(seg_flc_len(seg));
	  while (!(0 > (i))) {
	    long2str(tmp_str, 4*(i), seg_flc(seg)[i]);
	    {
	      i = -1+(i);
	    }
	  }
	}
	bt_put(seg_rt_han(seg), "FLC", 3, tmp_str, 4*(seg_flc_len(seg)));
	seg_set_flc_len(seg, -1);
      }
    {
      int ans = do_seg_buffers(seg, flush_buffer);
      if ((success_P(ans))
	  || (hammer))
	{
	  if (!(success_P(ans)))
	    ans = notpres;
	  do_seg_buffers(seg, purge_buffer);
	  bt_close(seg_rt_han(seg));
	  bt_close(seg_fl_han(seg));
	  blk_file_close(seg_port(seg));
	  seg_set_port(seg, 0);
	  seg_set_str(seg, 0);
	  seg_set_used(seg, 0);
	}
      return ans;
    }
  }
}


unsigned char db_version_str[] = "wb1a8";

unsigned char db_authors_str[] = "A. Jaffer, J. Finger, R. Zito-Wolf";

/*  */
/*  */
/* The integer @var{seg} must specify an unused segment.  The integer @var{bsiz} */
/* specifies the size of B-tree blocks.  Unless you are testing */
/* splitting algorithms, @var{bsiz} should be 2048. */
/*  */
/* @code{make_seg} makes a new empty database named @var{filename}.  A status code is returned. */
/* Call @code{open_seg} to use the new database. */

int make_seg(seg, filename, bsiz)
     int seg;
     unsigned char *filename;
     int bsiz;
{
  if ((!(seg_free_P(seg)))
      || (!(try_lck(seg_lck(seg)))))
    {
      dprintf((diagout, ">>>>ERROR<<<< %s: segment in use %d\n", "make_seg", seg));
      return argerr;
    }
  else if ((bsiz)>(blk_size))
    {
      dprintf((diagout, ">>>>ERROR<<<< unsupported %s %d > %d\n", "bsiz", bsiz, blk_size));
      unlck(seg_lck(seg));
      return argerr;
    }
  else {
    int file = blk_file_create(filename, bsiz);
    if (output_port_P(file))
      {
	seg_set_port(seg, file);
	seg_set_bsiz(seg, bsiz);
	seg_set_used(seg, 3);
	seg_set_str(seg, filename);
	init_leaf_blk(empty_blk, 0, dir_typ);
	blk_set_time(empty_blk, time(0L));
	blk_write(file, empty_blk, bsiz, 0);
	init_leaf_blk(empty_blk, 1, dir_typ);
	blk_set_time(empty_blk, time(0L));
	blk_write(file, empty_blk, bsiz, 1);
	init_leaf_blk(empty_blk, 2, frl_typ);
	blk_set_time(empty_blk, time(0L));
	blk_write(file, empty_blk, bsiz, 2);
	blk_file_close(file);
	file = blk_file_open_modify(filename, bsiz);
	if (output_port_P(file))
	  {
	    seg_set_port(seg, file);
	    unlck(seg_lck(seg));
	    {
	      HAND *han = seg_rt_han(seg);
	      unsigned char tmp_str[5];
	      bt_open(seg, 0, han, (wcb_sap)+(wcb_sar));
	      bt_put(han, "", 0, db_version_str, strlen(db_version_str));
	      long2str(tmp_str, 0, seg_used(seg));
	      bt_put(han, "USED", 4, tmp_str, 4);
	      short2str(tmp_str, 0, seg_bsiz(seg));
	      bt_put(han, "BSIZ", 4, tmp_str, 2);
	      tmp_str[0] = (4);
	      long2str(tmp_str, 1, 1);
	      bt_put(han, "ROOT", 4, tmp_str, 5);
	      long2str(tmp_str, 1, 2);
	      bt_put(han, "FLD", 3, tmp_str, 5);
	      bt_put(han, "FLC", 3, "", 0);
	      if ((bsiz)>0x80)
		bt_put(han, "authors", 7, db_authors_str, strlen(db_authors_str));
	      return close_seg(seg, 0);
	    }
	  }
	else {
	  dprintf((diagout,
	 ">>>>ERROR<<<< couldn't open fresh file %s\n",
	 filename));
	  unlck(seg_lck(seg));
	  return ioerr;
	}
      }
    else {
      dprintf((diagout,
	 ">>>>ERROR<<<< couldn't create new file %s\n",
	 filename));
      unlck(seg_lck(seg));
      return ioerr;
    }
  }
}

/* @noindent The write-control-bits argument (@var{wcb}) to these */
/* functions controls the latency of updates to the file after various */
/* operations.  These bits are defined as follows: */
/*  */
/* @multitable @columnfractions .1 .3 .6 */
/* @item @var{value} */
/* @tab C-name */
/* @tab Meaning */
/* @item 1 */
/* @tab wcb_sap */
/* @tab save block after PUTs */
/* @cindex wcb_sap */
/* @item 2 */
/* @tab wcb_sar */
/* @tab save block after REMOVEs */
/* @cindex wcb_sar */
/* @item 4 */
/* @tab wcb_sac */
/* @cindex wcb_sac */
/* @tab force block save after cached block changes */
/* @item 8 */
/* @tab wcb_fac */
/* @tab flush buffer entirely after cached block changes (not currently */
/* implemented) */
/* @cindex wcb_fac */
/* @end multitable */
/* Opens bt-handle @var{han} to seg number @var{seg}, block number @var{blk_num}, and */
/* returns the type of the block.  If no such block exists or is not a */
/* root block, then a (negative) status code is returned. */

int bt_open(seg, blk_num, han, wcb)
     int seg;
     long blk_num;
     HAND *han;
     int wcb;
{
  if (((seg)>=0)
      && ((seg)<(num_segs))
      && (seg_str(seg)))
    {
      ENTRY *ent = get_ent(seg, blk_num, accread);
      if (!(ent))
	return unkerr;
      else if (!(root_P(ent_blk(ent))))
	{
	  release_ent(ent, accread);
	  dprintf((diagout, ">>>>ERROR<<<< %s: not a root %d:%ld\n", "bt_open", seg, blk_num));
	  return argerr;
	}
      else {
	han_set_seg(han, seg);
	han_set_num(han, blk_num);
	han_set_typ(han, blk_typ(ent_blk(ent)));
	han_set_last(han, blk_num);
	if (blk_typ_P(ent_blk(ent), dir_typ))
	  wcb = (wcb)|((wcb_sap)+(wcb_sar));
	han_set_wcb(han, wcb);
	release_ent(ent, accread);
	return han_typ(han);
      }
    }
  else return argerr;
}

/* Creates a new root block in seg @var{seg} of type @var{typ}, opens bt-handle */
/* @var{han} to it, and returns a status code.  If @var{seg} has insufficient room to */
/* create a new tree, then the @var{noroom} status code is returned. */
/*  */
/* @code{bt_create} can be used to create temporary b-trees.  Temporary trees will be */
/* be reclaimed by check program after system crashes.  In order to */
/* make a tree persistent, add it to a directory (tree). */

int bt_create(seg, typ, han, wcb)
     int seg;
     int typ;
     HAND *han;
     int wcb;
{
  ENTRY *ent = create_new_blk_ent(seg);
  if (!(ent))
    return noroom;
  else {
    long blk_num = ent_id(ent);
    init_leaf_blk(ent_blk(ent), blk_num, typ);
    ent_set_dty(ent, !0);
    ent_set_pus(ent, 0);
    ent_write(ent);
    han_set_seg(han, seg);
    han_set_num(han, blk_num);
    han_set_typ(han, typ);
    han_set_last(han, blk_num);
    if ((typ)==(dir_typ))
      wcb = (wcb)|((wcb_sap)+(wcb_sar));
    han_set_wcb(han, wcb);
    release_ent(ent, accwrite);
    return success;
  }
}


/* Closes bt-handle @var{han} and returns @var{SUCCESS}. */
/*  */
/* Currently, @code{bt_close} has no effect other than to clear @var{han}. */

int bt_close(han)
     HAND *han;
{
  han_set_seg(han, 0);
  han_set_num(han, 0);
  han_set_typ(han, 0);
  han_set_last(han, 0);
  return success;
}

/* @section HANDs and Tree Operations */
/* @cindex HAND */

int clever_cache_enable = !0;

/* Note: most of the data-manipulating commands here */
/* can return @var{notpres}, with the followng meanings: */
/* @cindex notpres */
/*  */
/* @multitable @columnfractions .2 .8 */
/* @item @code{bt-get} */
/* @tab no such key. */
/* @item @code{bt-next} */
/* @tab no NEXT key (ie, key given was LAST key). */
/* @item @code{bt-prev} */
/* @tab no PREV key (ie, key given was FIRST key). */
/* @item @code{bt-rem} */
/* @tab KEY was not found. */
/* @c @item @code{bt-rem-range} */
/* @c @tab ?? */
/* @item @code{bt-put} */
/* @tab NOT USED (could be symmetric w/WRITE). */
/* @item @code{bt-write} */
/* @tab key WAS found, so no write done. */
/* @end multitable */
/* @var{key_str} is a string of length @var{k_len}.  @code{bt_get} stores into the string @var{ans_str} the */
/* value associated with @var{key_str} in tree @var{han}.  @code{bt_get} returns the length of the */
/* string stored into @var{ans_str} or an error code. */

int bt_get(han, key_str, k_len, ans_str)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *ans_str;
{
  int pkt[pkt_size];
  ENTRY *ent = 0;
  ent = chain_find_ent(han, accread, key_str, k_len, pkt);
  if (!(ent))
    {
      get_fct = 0x1L+(get_fct);
      return unkerr;
    }
  else if ((match_type(pkt))!=(match))
    {
      get_ct = 0x1L+(get_ct);
      release_ent(ent, accread);
      return notpres;
    }
  else {
    int alen = get_this_val(ent_blk(ent), match_pos(pkt), ans_str);
    get_ct = 0x1L+(get_ct);
    release_ent(ent, accread);
    return alen;
  }
}

/* @var{key_str} is a string of length @var{k_len}.  @code{bt_next} stores into the string @var{ans_str} the */
/* next key after @var{key_str} in tree @var{han}.  @code{bt_next} returns the length of the string */
/* stored into @var{ans_str} or an error code. */

int bt_next(han, key_str, k_len, ans_str)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *ans_str;
{
  int pkt[pkt_size];
  ENTRY *ent = 0;
  ent = chain_find_ent(han, accread, key_str, k_len, pkt);
  if (!(ent))
    {
      next_fct = 0x1L+(next_fct);
      return unkerr;
    }
  else {
    next_ct = 0x1L+(next_ct);
    {
      int res = chain_next(ent, key_str, k_len, ans_str, pkt);
      if (clever_cache_enable)
	han_set_last(han, blk_to_cache(pkt));
      return res;
    }
  }
}

/* @var{key_str} is a string of length @var{k_len}.  @code{bt_prev} stores into the string @var{ans_str} the */
/* last key before @var{key_str} in tree @var{han}.  @code{bt_prev} returns the length of the string */
/* stored into @var{ans_str} or an error code. */

int bt_prev(han, key_str, k_len, ans_str)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *ans_str;
{
  int pkt[pkt_size];
  ENTRY *ent = 0;
  ent = chain_find_prev_ent(han, accread, key_str, k_len, pkt);
  if (ent)
    ent = prev_k_ent(ent, key_str, k_len, leaf, pkt);
  if (!(ent))
    {
      prev_fct = 0x1L+(prev_fct);
      return unkerr;
    }
  else {
    prev_ct = 0x1L+(prev_ct);
    if (!(match_pos(pkt)))
      {
	release_ent(ent, accread);
	return notpres;
      }
    else {
      int k_len2 = recon_this_key(ent_blk(ent), match_pos(pkt), ans_str, 0, 0x100);
      han_set_last(han, ent_id(ent));
      release_ent(ent, accread);
      return k_len2;
    }
  }
}


/* @var{key_str} is a string of length @var{k_len}.  @code{bt_rem} stores into the string @var{ans_str} the */
/* value associated with @var{key_str} in tree @var{han}; then removes that association */
/* from tree @var{han}.  @code{bt_rem} returns the length of the string stored into @var{ans_str} or */
/* an error code. */
/*  */
/* If @var{ans_str} is 0, @code{bt_rem} removes the @var{key_str} association from tree @var{han} and returns */
/* @var{SUCCESS} if successful; an error code if not. */

int bt_rem(han, key_str, k_len, ans_str)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *ans_str;
{
  int pkt[pkt_size];
  ENTRY *ent = 0;
  if ((k_len)<0)
    {
      dprintf((diagout, ">>>>ERROR<<<< %s: bad length string %d\n", "bt_rem", k_len));
      return argerr;
    }
  else {
    ent = chain_find_ent(han, accwrite, key_str, k_len, pkt);
    if (ent)
      {
	rem_ct = 0x1L+(rem_ct);
	{
	  int ans = chain_rem(ent, key_str, k_len, ans_str, pkt, han_wcb(han));
	  release_ent(ent, accwrite);
	  return ans;
	}
      }
    else {
      rem_fct = 0x1L+(rem_fct);
      return unkerr;
    }
  }
}

/*  To make possible bounded-time operation, @0 will purge at most */
/*  BLK-LIMIT blocks at a time; passing BLK-LIMIT of -1 imposes no */
/*  limit. */

/* @var{key_str} must be a maximum-length (256 byte) string containing a key */
/* @var{k_len} bytes long.  @var{key2_str} is a string of length @var{k2_len}. */
/*  */
/* @code{bt_rem_range} removes [@var{key_str} @dots{} @var{key2_str}) and their values.  If @var{key2_str} <= @var{key_str} no */
/* deletion will occur (even if @var{key_str} is found).  @code{bt_rem_range} returns SUCCESS if */
/* the operation is complete, an error status code if not. */

int bt_rem_range(han, key_str, k_len, key2_str, k2_len)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *key2_str;
     int k2_len;
{
  int respkt[pkt_size];
  return bt_scan(han, rem_scan, key_str, k_len, key2_str, k2_len, 0, 0, respkt, -1);
}

/* @var{key_str} is a string of length @var{k_len}.  @var{val_str} is a string of length @var{v_len}.  @code{bt_put} */
/* makes the value associated with @var{key_str} be @var{val_str} in tree @var{han}.  @code{bt_put} returns a */
/* status code for the operation. */

int bt_put(han, key_str, k_len, val_str, v_len)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *val_str;
     int v_len;
{
  ENTRY *ent = 0;
  int pkt[pkt_size];
  if (((v_len)>0xff)
      || ((k_len)>0xff)
      || ((k_len)<0))
    return argerr;
  else {
    ent = chain_find_ent(han, accwrite, key_str, k_len, pkt);
    if (ent)
      {
	int res = chain_put(ent, key_str, k_len, val_str, v_len, pkt, 0, han_wcb(han));
	if (res)
	  {
	    if (clever_cache_enable)
	      han_set_last(han, blk_to_cache(pkt));
	    put_ct = 0x1L+(put_ct);
	    return success;
	  }
	else {
	  put_fct = 0x1L+(put_fct);
	  return unkerr;
	}
      }
    else return unkerr;
  }
}

/* @var{key_str} is a string of length @var{k_len}.  @var{val_str} is a string of length @var{v_len}.  If */
/* @var{han} currently contains an association for @var{key_str}, then @code{bt_write} does not modify */
/* the tree and returns the @var{notpres} status code. */
/*  */
/* Otherwise, @code{bt_write} makes the value associated with @var{key_str} be @var{val_str} in tree @var{han}. */
/* @code{bt_write} returns a status code for the operation. */

int bt_write(han, key_str, k_len, val_str, v_len)
     HAND *han;
     unsigned char *key_str;
     int k_len;
     unsigned char *val_str;
     int v_len;
{
  ENTRY *ent = 0;
  int pkt[pkt_size];
  if (((v_len)>0xff)
      || ((k_len)>0xff)
      || ((k_len)<0))
    return argerr;
  else {
    ent = chain_find_ent(han, accwrite, key_str, k_len, pkt);
    if (ent)
      if ((match_type(pkt))==(match))
	{
	  release_ent(ent, accwrite);
	  return notpres;
	}
      else {
	int res = chain_put(ent, key_str, k_len, val_str, v_len, pkt, 0, han_wcb(han));
	if (res)
	  {
	    if (clever_cache_enable)
	      han_set_last(han, blk_to_cache(pkt));
	    put_ct = 0x1L+(put_ct);
	    return success;
	  }
	else {
	  put_fct = 0x1L+(put_fct);
	  return unkerr;
	}
      }
    else return unkerr;
  }
}
